;       ----------------------------------------------------

;    scrollNES - version 0.1
;    Copyright 2010 Don Miller & Andrew Reitano
;    For more information, visit: http://www.no-carrier.com

;    This program is free software: you can redistribute it AND/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation, either version 3 of the License, or
;    (at your option) any later version.

;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.

;    You should have received a copy of the GNU General Public License
;    along with this program.  If not, see <http://www.gnu.org/licenses/>.

;       ----------------------------------------------------

NewButtons = $41
OldButtons = $42
JustPressed = $43
PaletteNumber = $46
scroll_h = $c0
scroll_v = $c1
speed = $c2
offset_counter = $c3
letter_counter = $c4
char_counter = $c5
char_address = $c6              ; Points to beginning of desired character memory
write_width = $c8
size_toggle = $d0               ; 8x8 or 16x8
char_val = $d1                  ; tile 0,1,2,3
size = $d2
char = $d3
speed_toggle = $d4
sine_counter = $d5
string_pos = $d6                ; (and $d7) Keeps track of which character we are writing on the message

;       ----------------------------------------------------

        .ORG $7ff0
Header:                         ; 16 byte .NES header (iNES format)
	.DB "NES", $1a
	.DB $02                 ; size of PRG ROM in 16kb units
	.DB $01			; size of CHR ROM in 8kb units
	.DB #%00000000		; mapper 0
	.DB #%00000000		; mapper 0
        .DB $00
        .DB $00
        .DB $00
        .DB $00
        .DB $00
        .DB $00
        .DB $00
        .DB $00

        .ORG $8000

;       ----------------------------------------------------

Reset:                          ; reset routine
        SEI
        CLD
	LDX #$00
	STX $2000
	STX $2001
	DEX
	TXS
  	LDX #0
  	TXA
ClearMemory:
	STA 0, X
	STA $100, X
	STA $200, X
	STA $300, X
	STA $400, X
	STA $500, X
	STA $600, X
	STA $700, X
        STA $800, X
        STA $900, X
        INX
	BNE ClearMemory

;       ----------------------------------------------------

        LDA #$00                ; setting up variables
        STA PaletteNumber
        STA scroll_h
        STA scroll_v
        STA offset_counter
        STA letter_counter
        STA char_counter
        STA write_width

        STA char_val
        STA char
        STA size_toggle
        STA size
        STA sine_counter

        LDA #$01
        STA speed
        STA speed_toggle

	LDA #<string
	STA string_pos+0
	LDA #>string
	STA string_pos+1

;       ----------------------------------------------------

	LDX #$02                ; warm up
WarmUp:
	BIT $2002
	BPL WarmUp
	DEX
	BNE WarmUp

       	LDA #$3F
	STA $2006
	LDA #$00
	STA $2006
        TAX
LoadPal:                        ; load palette
        LDA palette, x
        STA $2007
        INX
        CPX #$10
        BNE LoadPal

	LDA #$20
	STA $2006
	LDA #$00
	STA $2006

	LDY #$04                ; clear nametables
ClearName:
	LDX #$00
	LDA #$00
PPULoop:
	STA $2007
	DEX
	BNE PPULoop

	DEY
	BNE ClearName

;       ----------------------------------------------------

        JSR LoadNewPalette
        JSR InitSprites
        JSR Vblank              ; turn on screen

;       ----------------------------------------------------

InfLoop:                        ; loop forever, program now controlled by NMI routine

        JMP InfLoop

;       ----------------------------------------------------

LoadNewPalette:
       	LDX PaletteNumber       ; load palette lookup value
        LDY #$00
        LDA #$3F
	STA $2006
	LDA #$00
	STA $2006
LoadNewPal:                     ; load palette
        LDA palette, x
        STA $2007
        INX
        INY
        CPY #$20
        BNE LoadNewPal
        RTS

;       ----------------------------------------------------

Vblank:                         ; turn on the screen
	BIT $2002
	BPL Vblank

        LDX scroll_h
        STX $2005
        LDX scroll_v
        STX $2005

	LDA #%10101100
	STA $2000
        
        LDA #%00011110
	STA $2001

        RTS

;       ----------------------------------------------------

ControllerTest:

        LDA NewButtons
	STA OldButtons

        LDX #$00
	LDA #$01		; strobe joypad
	STA $4016
	LDA #$00
	STA $4016
ConLoop:
	LDA $4016		; check the state of each button
	LSR
	ROR NewButtons
        INX
        CPX #$08
        BNE ConLoop

	LDA OldButtons          ; invert bits
	EOR #$FF
	AND NewButtons
	STA JustPressed

CheckSelect:
	LDA #%00000100
	AND JustPressed
	BEQ CheckStart

	LDA size_toggle
	eor #$01
	STA size_toggle

CheckStart:
	LDA #%00001000
	AND JustPressed
	BEQ CheckB

	INC char_val
        INC char_val
	LDA char_val
	cmp #8
	BNE CheckB
	LDA #0
	STA char_val

CheckB:
	LDA #%01000010
	AND JustPressed
	BEQ CheckA

        LDA speed_toggle
        eor #$01
        STA speed_toggle

CheckA:
	LDA #%10000001
	AND JustPressed
        BEQ CheckOver

CheckOver:

        RTS

;       ----------------------------------------------------

InitSprites:
      LDA #$ff
      LDX #$00
ClearSprites:
      STA $500, x
      INX
      BNE ClearSprites

      LDA #$00
      STA $2003                 ; set the low byte (00) of the RAM address
      LDA #$05
      STA $4014                 ; set the high byte (05) of the RAM address

LoadSprites:
      LDX #$00
LoadSpritesLoop:
      LDA sprites, x            ; load data from address
      STA $0500, x              ; store into RAM address
      INX
      CPX #196
      BNE LoadSpritesLoop
      RTS

;       ----------------------------------------------------

UpdateSprites:
        LDA #$00
        STA $2003
        LDA #$05
        STA $4014
        RTS

;       ----------------------------------------------------

sprites:
           ;vert tile attr horiz
        .db $37, $30, $00, $00  ; sprite rows (left)
        .db $47, $30, $00, $00
        .db $57, $30, $00, $00
        .db $67, $30, $00, $00
        .db $77, $30, $00, $00
        .db $87, $30, $00, $00
        .db $97, $30, $00, $00
        .db $A7, $30, $00, $00

        .db $37, $32, $00, $08
        .db $47, $32, $00, $08
        .db $57, $32, $00, $08
        .db $67, $32, $00, $08
        .db $77, $32, $00, $08
        .db $77, $32, $00, $08
        .db $87, $32, $00, $08
        .db $97, $32, $00, $08
        .db $A7, $32, $00, $08

        .db $37, $34, $00, $10
        .db $47, $34, $00, $10
        .db $57, $34, $00, $10
        .db $67, $34, $00, $10
        .db $77, $34, $00, $10
        .db $87, $34, $00, $10
        .db $97, $34, $00, $10
        .db $A7, $34, $00, $10

        .db $37, $34, $00, $e8  ; sprite rows (right)
        .db $47, $34, $00, $e8
        .db $57, $34, $00, $e8
        .db $67, $34, $00, $e8
        .db $77, $34, $00, $e8
        .db $87, $34, $00, $e8
        .db $97, $34, $00, $e8
        .db $A7, $34, $00, $e8
      
        .db $37, $32, $00, $f0
        .db $47, $32, $00, $f0
        .db $57, $32, $00, $f0
        .db $67, $32, $00, $f0
        .db $77, $32, $00, $f0
        .db $87, $32, $00, $f0
        .db $97, $32, $00, $f0
        .db $A7, $32, $00, $f0
      
        .db $37, $30, $00, $f8
        .db $47, $30, $00, $f8
        .db $57, $30, $00, $f8
        .db $67, $30, $00, $f8
        .db $77, $30, $00, $f8
        .db $87, $30, $00, $f8
        .db $97, $30, $00, $f8
        .db $A7, $30, $00, $f8

;       ----------------------------------------------------

DrawLoop:                       ;** CALCULATE CHARACTER ADDRESS **
       	LDY #$00
	LDA (string_pos), y     ; Determine which character on string to load

        BNE ContinueStr         ; Check if we've terminated the string, if we haven't - leave it alone

	LDA #<string            ; Otherwise reset position
	STA string_pos+0
	LDA #>string
	STA string_pos+1
       	LDY #$00
	LDA (string_pos), y

;       ----------------------------------------------------

        LDX size_toggle        ; Change size, speed and character set after string loops
        STX size
        LDX char_val
        STX char
        LDX speed_toggle
        STX speed

;       ----------------------------------------------------

ContinueStr:
        SEC                     ; Clear carry ($20-$20 could result in $FF!)
        SBC #$20                ; Our ASCII starts at 32, but our character map starts at zero, subtract this offset
        ; First take care of the HI address
        ; 7 6 5 4 3 2 1 0
        ; x x x x x x L L - Since the two L's will repeat 4 times before "overflowing" (4*64) we can assume..
        ;
        ; M M M M M M x x - that the remaining bits represent how much to add to the HI address
        ; Shift them right 2 and add them to the HI (ex. $C0 -> $C1)
        LSR                     ; Perform shifts
        LSR
        CLC
        ADC #$C0                ; Add to our offset (second bank)
        STA char_address+1

                                ; Take care of LO address
      	LDY #$00                ; Determine which character on string to load
	LDA (string_pos), y
        SEC                     ; Need to set carry here or we borrow a bit
        SBC #$20                ; Subtract ASCII offset (32)
        ASL                     ; Multiply by 64 (Shift left 6 times)
        ASL
        ASL
        ASL
        ASL
        ASL
        STA char_address+0

        ; Final result $C000 -> $C040 -> $C080 -> $C0C0 -> $C100 AND so on

;       ----------------------------------------------------

        LDX offset_counter
        LDY offset_table,x

        LDA #$20
        STA $2006
        STY $2006

        LDX #8
ReadChar:
        LDY char_counter
        LDA (char_address),y    ; Use post-address indexing to use full memory address

        CLC
        ADC char
        STA $2007
        TAY

        LDA size                ; Check status of size variable to decide if there should be an extra write
        BEQ small
        STY $2007
small:
        INC char_counter
        DEX
        BNE ReadChar

        LDA char_counter        ; Do not exceed 64 (8x8->length of character)
        AND #$3f
        STA char_counter
        BNE SkipInc             ; While we're here, Increment the string position if the result is zero

	CLC
        LDA string_pos+0        ;increment pointer in text loop
	ADC #1
	STA string_pos+0
        LDA string_pos+1
	ADC #0
	STA string_pos+1

SkipInc:
        RTS

;       ----------------------------------------------------

EraseInit:
        INC offset_counter
        LDA offset_counter
        CMP #32
        BNE EraseColumn
        LDA #0
        STA offset_counter

EraseColumn:
        LDX offset_counter
        LDY offset_table,x      ; Use offset table to calculate screen position

        LDA #$20
        STA $2006
        STY $2006

        LDX #16
EraseLoop:
        LDA #$00                ; Write blank tile (erase)
        STA $2007
        DEX                     ; 16x to cover height of character
        BNE EraseLoop

        RTS

;       ----------------------------------------------------

NMI:

        INC scroll_h            ; Update scroll counters
        INC scroll_h
        LDA speed
        BEQ SlowScroll
        INC scroll_h
        INC scroll_h

SlowScroll:
        LDA scroll_h
        AND #$07
        BNE KeepGoing

        JSR DrawLoop
        JSR EraseInit

KeepGoing:
        LDX scroll_h            ; Update scroll registers
        STX $2005
        LDX scroll_v
        STX $2005

        JSR UpdateSprites       ; Comment this line out to disable sprites
        JSR ControllerTest      ; check for user input

        RTI
IRQ:
        RTI

;       ----------------------------------------------------

offset_table:
        .byte $e0,$e1,$e2,$e3,$e4,$e5,$e6,$e7,$e8,$e9,$ea,$eb,$ec,$ed,$ee,$ef
        .byte $f0,$f1,$f2,$f3,$f4,$f5,$f6,$f7,$f8,$f9,$fa,$fb,$fc,$fd,$fe,$ff

;       ----------------------------------------------------

palette:                        ; palette data
        .byte $0F,$30,$20,$00,$0F,$20,$20,$20,$0F,$10,$10,$10,$0F,$00,$00,$00
        .byte $0F,$0f,$0f,$0f,$0F,$20,$20,$20,$0F,$10,$10,$10,$0F,$00,$00,$00

;       ----------------------------------------------------

string:
        .db "Happy New Year 2010 from NO CARRIER & Batsly Adams!     ",$00

;       ----------------------------------------------------

        .ORG $C000
        .include "charmap.bin"    ; Attach character map to ROM

	.ORG $fffa
	.DW NMI
	.DW Reset
	.DW IRQ